from functools import cache

import polars as pl
from rdkit import Chem
from tqdm import tqdm

from .polars import w_pbar


@cache
def canonical_smiles_wo_salt(smiles: str) -> str | None:
    """
    Get the canonical SMILES without salt from the input SMILES.

    Salt is a short part separated by "." in the SMILES.
    Shared function with dti-pytorch
    """
    m = Chem.MolFromSmiles(smiles)
    if m is not None:
        canonical_smiles = Chem.MolToSmiles(m, isomericSmiles=True, canonical=True)
        split_smi = canonical_smiles.split(".")
        if len(split_smi) > 1:
            smiles_wo_salt = max(split_smi, key=len)
            if Chem.MolFromSmiles(smiles_wo_salt) is None:
                smiles_wo_salt = None
        else:
            smiles_wo_salt = split_smi[0]
    else:
        smiles_wo_salt = None
    return smiles_wo_salt


def polars_canonical_smiles_wo_salt(
    df: pl.DataFrame,
    *,
    smiles_col: str = "smiles",
    out_col: str = "canonical_smiles_wo_salt",
) -> pl.DataFrame:
    """
    Apply canonical_smiles_wo_salt on the DataFrame with tqdm.
    """
    with tqdm(
        total=df.shape[0], desc="Converting smiles to canonical smiles without salt"
    ) as pbar:
        df = df.with_columns(
            pl.col(smiles_col)
            .map_elements(w_pbar(pbar, canonical_smiles_wo_salt), return_dtype=pl.Utf8)
            .alias(out_col),
        )

    return df



def uniprot_xml_to_postgresql(
    *,
    uniprot_xml_path: str,
    uri: str,
) -> None:
    """
    (🦀 Rust) Load UniProt XML file into PostgreSQL database.

    This creates a `uniprot` database and a `uniprot_info` table.
    """



import html

import polars as pl
import sqlalchemy


def fix_assay_table(uri: str):
    """
    Fix the assay table in MySQL by decoding HTML entities like `&#39;` and strip empty spaces.

    Notes:
        - the table is replaced.
        - types are reserved by manually changing them to the original types. For example, varchar -> text -> varchar
        - primary key and foreign key constraints are reserved by manually adding them back just like the original table
    """
    query = """
        SELECT
          *
        FROM
          assay
    """
    assay_df = pl.read_database_uri(query=query, uri=uri)

    # the column might be "binary" type if the type is "TEXT" in MySQL
    assay_df = assay_df.with_columns(
        pl.col("description")
        .cast(pl.Utf8)
        .map_elements(lambda s: html.unescape(s.strip()), return_dtype=pl.Utf8),
        pl.col("assay_name")
        .cast(pl.Utf8)
        .map_elements(lambda s: html.unescape(s.strip()), return_dtype=pl.Utf8),
    )

    assay_df.write_database(
        table_name="assay",
        connection=uri,
        if_table_exists="replace",
    )

    with sqlalchemy.create_engine(uri).connect() as conn:
        conn.execute(
            sqlalchemy.text("""
                ALTER TABLE assay MODIFY COLUMN `entryid` INT(11);
                ALTER TABLE assay MODIFY COLUMN `assayid` INT(11);
                ALTER TABLE assay MODIFY COLUMN `description` VARCHAR(4000);
                ALTER TABLE assay MODIFY COLUMN `assay_name` VARCHAR(200);
                ALTER TABLE assay ADD PRIMARY KEY (`entryid`,`assayid`);
                ALTER TABLE assay ADD CONSTRAINT `assay_ibfk_1` FOREIGN KEY (`entryid`) REFERENCES `entry` (`entryid`);
            """)
        )


# if __name__ == "__main__":
#     fix_assay_table("mysql://username:@localhost:3306/bind")



import polars as pl


def polars_standardize_fasta(
    df: pl.DataFrame, fasta_col: str = "fasta", out_col: str = "fasta"
) -> pl.DataFrame:
    """
    Remove spaces and make it uppercase of a Polars column.
    """
    df = df.with_columns(
        pl.col(fasta_col)
        .str.to_uppercase()
        .str.replace_all("\n", "")
        .str.replace_all(" ", "")
        .alias(out_col)
    )

    return df



import typer

from . import bindingdb, uniprot

app = typer.Typer(no_args_is_help=True)

app.add_typer(uniprot.app, name="uniprot")
app.add_typer(bindingdb.app, name="bindingdb")


def main():
    app()


if __name__ == "__main__":
    main()



import logging
from os import PathLike

import polars as pl
import psycopg

from bio_data_to_db.utils.postgresql import (
    create_db_if_not_exists,
    create_schema_if_not_exists,
    polars_write_database,
)

logger = logging.getLogger(__name__)


def create_empty_table(
    uri: str,
):
    """
    Create an empty table in the database. Necessary to create the table structure before inserting data.

    Note:
        It runs the following SQL query:
        ```sql
        CREATE TABLE public.uniprot_info (
          uniprot_pk_id BIGINT GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY,
          accessions TEXT[],
          names TEXT[],
          protein_names TEXT[],
          gene_names TEXT[],
          organism_scientific TEXT,
          organism_commons TEXT[],
          organism_synonyms TEXT[],
          ncbi_taxonomy_id INT,
          deargen_ncbi_taxonomy_id INT,
          lineage TEXT[],
          keywords TEXT[],
          geneontology_ids TEXT[],
          geneontology_names TEXT[],
          sequence TEXT,
          deargen_molecular_functions TEXT[]
        )
        ```
    """
    uri_wo_dbname, dbname = uri.rsplit("/", 1)
    create_db_if_not_exists(uri_wo_dbname, dbname)
    create_schema_if_not_exists(uri, "public")

    with psycopg.connect(
        conninfo=uri,
    ) as conn:
        try:
            cursor = conn.cursor()
            conn.autocommit = True
            cursor.execute(
                query="""
                CREATE TABLE public.uniprot_info (
                  uniprot_pk_id BIGINT GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY,
                  accessions TEXT[],
                  names TEXT[],
                  protein_names TEXT[],
                  gene_names TEXT[],
                  organism_scientific TEXT,
                  organism_commons TEXT[],
                  organism_synonyms TEXT[],
                  ncbi_taxonomy_id INT,
                  deargen_ncbi_taxonomy_id INT,
                  lineage TEXT[],
                  keywords TEXT[],
                  geneontology_ids TEXT[],
                  geneontology_names TEXT[],
                  sequence TEXT,
                  deargen_molecular_functions TEXT[]
                )
            """
            )
            logger.info(f"Database '{dbname}.public.uniprot_info' created successfully")

        except psycopg.Error:
            logger.exception(f"Error creating database '{dbname}.public.uniprot_info'")


def create_accession_to_pk_id(uri: str):
    """
    Create a table to map accession to uniprot_pk_id, from the uniprot_info table.

    It creates the following tables:

    - accession_to_pk_id
    - accession_to_pk_id_list

    Note:
        The mapping is not unique. It is possible to have multiple uniprot_pk_id for a single accession and vice versa.
    """
    with psycopg.connect(
        conninfo=uri,
    ) as conn:
        try:
            cursor = conn.cursor()
            conn.autocommit = True
            cursor.execute(
                query="""
                CREATE TABLE public.accession_to_pk_id (
                  accession TEXT,
                  uniprot_pk_id BIGINT
                )
            """
            )
            logger.info(
                "Table structure 'uniprot.public.accession_to_pk_id' created successfully"
            )

            cursor.execute(
                query="""
                INSERT INTO public.accession_to_pk_id (accession, uniprot_pk_id)
                SELECT UNNEST(accessions), uniprot_pk_id
                FROM public.uniprot_info
            """
            )
            logger.info(
                "Table 'uniprot.public.accession_to_pk_id' insert content successfully"
            )

            cursor.execute(
                query="""
                CREATE TABLE public.accession_to_pk_id_list (
                  accession TEXT PRIMARY KEY,
                  uniprot_pk_ids BIGINT[]
                )
            """
            )
            logger.info(
                "Table structure 'uniprot.public.accession_to_pk_id_list' created successfully"
            )

            cursor.execute(
                query="""
                INSERT INTO public.accession_to_pk_id_list (accession, uniprot_pk_ids)
                SELECT accession, ARRAY_AGG(uniprot_pk_id) AS uniprot_pk_ids
                FROM public.accession_to_pk_id
                GROUP BY accession;
            """
            )
            logger.info(
                "Table 'uniprot.public.accession_to_pk_id_list' content added successfully"
            )

        except psycopg.Error:
            logger.exception("Error creating table 'uniprot.public.accession_to_pk_id'")


def keywords_tsv_to_postgresql(
    keywords_tsv_file: str | PathLike,
    uri: str,
    schema_name="public",
    table_name="keywords",
):
    """
    Load the keywords_all_2024_06_26.tsv (or similar version) file into the database.
    """
    tsv_columns = [
        "Keyword ID",
        "Name",
        "Category",
        "Gene Ontologies",
    ]
    polars_columns = [
        "keyword_id",
        "name",
        "category",
        "geneontology_names",
    ]
    rename_dict = dict(zip(tsv_columns, polars_columns, strict=True))
    polars_dtypes = [
        pl.Utf8,
        pl.Utf8,
        pl.Utf8,
        pl.Utf8,
    ]

    df = pl.read_csv(
        str(keywords_tsv_file),
        separator="\t",
        columns=tsv_columns,
        # NOTE: Do not use new_columns with columns. It will result in wrong data.
        # new_columns=polars_columns,
        schema_overrides=polars_dtypes,
    )

    df = df.rename(rename_dict)

    # geneontology_names is a list of strings
    df = df.with_columns(
        geneontology_names=pl.col("geneontology_names")
        .str.split(", ")
        .cast(pl.List(pl.Utf8))
    )

    polars_write_database(
        df, schema_name=schema_name, table_name=table_name, connection=uri
    )



from .. import uniprot_xml_to_postgresql
from .utils import (
    create_accession_to_pk_id,
    create_empty_table,
    keywords_tsv_to_postgresql,
)

__all__ = [
    "uniprot_xml_to_postgresql",
    "create_empty_table",
    "create_accession_to_pk_id",
    "keywords_tsv_to_postgresql",
]



from typing import Annotated

import typer

app = typer.Typer(no_args_is_help=True)


@app.command(no_args_is_help=True)
def create_empty_table(
    uri: Annotated[
        str,
        typer.Argument(help="URI to the PostgreSQL database"),
    ],
):
    from bio_data_to_db.uniprot import create_empty_table
    from bio_data_to_db.utils.log import setup_logging

    setup_logging()
    create_empty_table(uri)


@app.command(no_args_is_help=True)
def xml_to_postgresql(
    uniprot_xml_file: Annotated[
        str, typer.Argument(help="Path to the UniProt XML file")
    ],
    uri: Annotated[str, typer.Argument(help="URI to the PostgreSQL database")],
):
    """
    Convert a UniProt XML file to a PostgreSQL database.

    You must first create the table using the `create_empty_table` command.
    """
    from bio_data_to_db.uniprot import uniprot_xml_to_postgresql
    from bio_data_to_db.utils.log import setup_logging

    setup_logging()
    uniprot_xml_to_postgresql(uniprot_xml_path=uniprot_xml_file, uri=uri)


@app.command(no_args_is_help=True)
def create_accession_to_pk_id(
    uri: Annotated[
        str,
        typer.Argument(help="URI to the PostgreSQL database"),
    ],
):
    """
    Create a dictionary mapping UniProt accession numbers to primary key IDs.

    Note that it is not unique, as the same accession number can be associated with multiple primary key IDs.
    """
    from bio_data_to_db.uniprot import create_accession_to_pk_id
    from bio_data_to_db.utils.log import setup_logging

    setup_logging()
    create_accession_to_pk_id(uri)


@app.command(no_args_is_help=True)
def keywords_tsv_to_postgresql(
    keywords_tsv_file: Annotated[
        str, typer.Argument(help="Path to the UniProt keywords TSV file")
    ],
    uri: Annotated[
        str,
        typer.Argument(help="URI to the PostgreSQL database"),
    ],
):
    """
    Convert a UniProt keywords TSV file to a PostgreSQL database.
    """
    from bio_data_to_db.uniprot import keywords_tsv_to_postgresql
    from bio_data_to_db.utils.log import setup_logging

    setup_logging()
    keywords_tsv_to_postgresql(keywords_tsv_file=keywords_tsv_file, uri=uri)


def main():
    app()


if __name__ == "__main__":
    main()



import enum
import logging
from typing import Annotated

import typer

logger = logging.getLogger(__name__)

app = typer.Typer(no_args_is_help=True)


class FixTableOption(str, enum.Enum):
    assay = "assay"


@app.command(no_args_is_help=True)
def fix_table(
    table_name: Annotated[
        FixTableOption,
        typer.Argument(help="Table name to fix"),
    ],
    uri: Annotated[
        str,
        typer.Argument(help="URI to the MySQL database"),
    ],
):
    """
    Fix the assay table in MySQL by decoding HTML entities like '&#39;' and strip empty spaces.
    """
    from bio_data_to_db.bindingdb.fix_tables import fix_assay_table
    from bio_data_to_db.utils.log import setup_logging

    setup_logging()
    fix_assay_table(uri)
    logger.info(
        "In `assay` table, HTML entities are decoded and empty spaces are stripped."
    )


def main():
    app()


if __name__ == "__main__":
    main()



import logging
import os
from pathlib import Path

from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

logger = logging.getLogger(__name__)

console = Console(
    theme=Theme(
        {
            "logging.level.error": "bold red blink",
            "logging.level.critical": "red blink",
            "logging.level.warning": "yellow",
        }
    )
)


def setup_logging(
    console_level: int | str = logging.INFO,
    output_files: list[str] | None = None,
    file_levels: list[int | str] | None = None,
):
    """
    Setup logging with RichHandler and FileHandler.

    You should call this function at the beginning of your script.

    Args:
        console_level: Logging level for console. Defaults to INFO or env var PPMI_LOG_LEVEL.
        output_files: List of output file paths, relative to LOG_DIR. If None, use default.
        file_levels: List of logging levels for each output file. If None, use default.
    """
    if output_files is None:
        output_files = []
    if file_levels is None:
        file_levels = []

    assert len(output_files) == len(
        file_levels
    ), "output_files and file_levels must have the same length"

    # NOTE: Initialise with NOTSET level and null device, and add stream handler separately.
    # This way, the root logging level is NOTSET (log all), and we can customise each handler's behaviour.
    # If we set the level during the initialisation, it will affect to ALL streams,
    # so the file stream cannot be more verbose (lower level) than the console stream.
    logging.basicConfig(
        format="",
        level=logging.NOTSET,
        stream=open(os.devnull, "w"),  # noqa: SIM115
    )

    # If you want to suppress logs from other modules, set their level to WARNING or higher
    # logging.getLogger("slowfast.utils.checkpoint").setLevel(logging.WARNING)

    console_handler = RichHandler(
        level=console_level,
        show_time=True,
        show_level=True,
        show_path=True,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        console=console,
    )
    console_format = logging.Formatter(
        fmt="%(name)s - %(message)s",
        datefmt="%m/%d %H:%M:%S",
    )
    console_handler.setFormatter(console_format)

    f_format = logging.Formatter(
        fmt="%(asctime)s - %(name)s: %(lineno)4d - %(levelname)s - %(message)s",
        datefmt="%y/%m/%d %H:%M:%S",
    )

    root_logger = logging.getLogger()
    root_logger.addHandler(console_handler)

    log_paths = []
    for output_file, file_level in zip(output_files, file_levels, strict=True):
        Path(output_file).parent.mkdir(parents=True, exist_ok=True)

        f_handler = logging.FileHandler(output_file)
        f_handler.setLevel(file_level)
        f_handler.setFormatter(f_format)

        # Add handlers to the logger
        root_logger.addHandler(f_handler)

    for log_path in log_paths:
        logger.info(f"Logging to {log_path}")


def main():
    logger.info("Hello, world!")
    raise Exception("Test exception")


if __name__ == "__main__":
    try:
        setup_logging()
        main()
    except Exception:
        logger.exception("Exception occurred")



from collections.abc import Callable
from typing import Any

import tqdm


def w_pbar(pbar: tqdm.std.tqdm, func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Apply progress bar when using `map_elements` in `polars`.

    Examples:
        >>> with tqdm(total=len(df)) as pbar:  # doctest: +SKIP
        ...     df = df.with_columns(
        ...         pl.col("in_col")
        ...         .map_elements(w_pbar(pbar, lambda x: x + 1), return_dtype=pl.Int64)
        ...     )

    Reference:
        - https://stackoverflow.com/questions/75550124/python-polars-how-to-add-a-progress-bars-to-apply-loops
    """

    def foo(*args, **kwargs):
        pbar.update(1)
        return func(*args, **kwargs)

    return foo



import logging
from collections.abc import Sequence

import polars as pl
import psycopg
from polars._typing import DbWriteMode
from psycopg import sql
from sqlalchemy import Engine, create_engine, types

logger = logging.getLogger(__name__)


# def polars_datatype_to_postgresql_type(dtype: pl.datatypes.DataType) -> str:
#     match dtype:
#         case pl.Utf8:
#             pg_element_type = "text"
#         case pl.UInt64:
#             # WARN: ignore unsigned because PostgreSQL doesn't support unsigned
#             pg_element_type = "bigint"
#         case pl.Int64:
#             pg_element_type = "bigint"
#         case pl.Int32:
#             pg_element_type = "integer"
#         case pl.Int16:
#             pg_element_type = "smallint"
#         case pl.Float64:
#             pg_element_type = "double precision"
#         case pl.Float32:
#             pg_element_type = "real"
#         case _:
#             raise ValueError(f"Unsupported inner dtype: {dtype}")
#
#     return pg_element_type


def polars_datatype_to_sqlalchemy_type(
    dtype: pl.datatypes.DataType,
) -> types.TypeEngine:
    match dtype:
        case pl.Utf8:
            sqlalchemy_type = types.TEXT()
        case pl.UInt64:
            # WARN: ignore unsigned because PostgreSQL doesn't support unsigned. Possible loss of data
            sqlalchemy_type = types.BIGINT()
        case pl.Int64:
            sqlalchemy_type = types.BIGINT()
        case pl.UInt32:
            sqlalchemy_type = types.BIGINT()
        case pl.Int32:
            sqlalchemy_type = types.INTEGER()
        case pl.UInt16:
            sqlalchemy_type = types.INTEGER()
        case pl.Int16:
            sqlalchemy_type = types.SMALLINT()
        case pl.Float64:
            sqlalchemy_type = types.DOUBLE_PRECISION()
        case pl.Float32:
            sqlalchemy_type = types.REAL()
        case pl.Boolean:
            sqlalchemy_type = types.BOOLEAN()
        case pl.List:
            inner_dtype = dtype.inner
            assert inner_dtype is not None
            inner_sqlalchemy_type = polars_datatype_to_sqlalchemy_type(inner_dtype)
            sqlalchemy_type = types.ARRAY(inner_sqlalchemy_type)
        case _:
            raise ValueError(f"Unsupported dtype: {dtype}")

    return sqlalchemy_type


def create_db_if_not_exists(uri_wo_db: str, db_name: str, comment: str | None = None):
    """
    Create a database if it doesn't exist.
    """
    with psycopg.connect(
        conninfo=f"{uri_wo_db}",
    ) as conn:
        try:
            cursor = conn.cursor()
            conn.autocommit = True
            cursor.execute(
                query=sql.SQL("""CREATE DATABASE {db_name};""").format(
                    db_name=sql.Identifier(db_name)
                )
            )
            if comment is not None:
                cursor.execute(
                    query=sql.SQL(
                        """COMMENT ON DATABASE {db_name} IS {comment};"""
                    ).format(
                        db_name=sql.Identifier(db_name),
                        comment=sql.Literal(comment),
                    )
                )
            logger.info(f"Database '{db_name}' created successfully")
        except psycopg.errors.DuplicateDatabase:
            logger.info(f"Database '{db_name}' already exists, Skip creating database.")
            if comment is not None:
                conn.rollback()
                cursor = conn.cursor()
                cursor.execute(
                    query=sql.SQL(
                        """COMMENT ON DATABASE {db_name} IS {comment};"""
                    ).format(
                        db_name=sql.Identifier(db_name),
                        comment=sql.Literal(comment),
                    )
                )

        except psycopg.Error:
            logger.exception(f"Error creating database '{db_name}'")


def create_schema_if_not_exists(uri: str, schema_name: str, comment: str | None = None):
    """
    Create a schema if it doesn't exist. The DB should already exist.
    """
    db_name = uri.split("/")[-1]
    with psycopg.connect(
        conninfo=uri,
    ) as conn:
        try:
            cursor = conn.cursor()
            cursor.execute(
                query=sql.SQL("""CREATE SCHEMA {schema_name};""").format(
                    schema_name=sql.Identifier(schema_name)
                )
            )
            if comment is not None:
                cursor.execute(
                    query=sql.SQL(
                        """COMMENT ON SCHEMA {schema_name} IS {comment};"""
                    ).format(
                        schema_name=sql.Identifier(schema_name),
                        comment=sql.Literal(comment),
                    )
                )
            conn.commit()

            logger.info(
                f"Schema '{schema_name}' created successfully in DB '{db_name}'"
            )
        except psycopg.errors.DuplicateSchema:
            logger.info(
                f"Schema '{schema_name}' in DB '{db_name}' already exists, Skip creating schema."
            )
            if comment is not None:
                # cancel the transaction and try to add comment
                conn.rollback()
                cursor = conn.cursor()
                cursor.execute(
                    query=sql.SQL(
                        """COMMENT ON SCHEMA {schema_name} IS {comment};"""
                    ).format(
                        schema_name=sql.Identifier(schema_name),
                        comment=sql.Literal(comment),
                    )
                )
                conn.commit()
        except psycopg.Error:
            logger.exception(f"Error creating schema '{schema_name}' in DB '{db_name}'")


def make_int_column_primary_key_identity(
    uri: str,
    *,
    schema_name: str = "public",
    table_name: str,
    column_name: str = "index",
):
    """
    Make an existing index column (integer type) as primary key with auto increment (identity).

    This is used because pl.DataFrame.write_database() doesn't support writing index column as primary key.
    Also, it will automatically set the start value of auto increment to the max value in the column.

    Example:
        >>> df = pl.DataFrame({"smiles": ["CCO", "CCN", "CCC"]})  # doctest: +SKIP
        ... df = df.with_row_index("pk_id")
        ... df.write_database(...)
        ... set_column_as_primary_key(uri=uri, table_name="table", column_name="pk_id")
        ... df2 = pl.DataFrame({"smiles": ["CCC", "CCN", "CCO"]})  # append without pk_id
        ... df2.write_database(..., if_table_exists="append")  # it will auto increment pk_id
    """
    with psycopg.connect(
        conninfo=uri,
    ) as conn:
        try:
            cursor = conn.cursor()
            # NOTE: since there are already values in the column, we need to set the start value to max+1

            cursor.execute(
                sql.SQL("""
            SELECT MAX({column}) FROM {table}
            """).format(
                    column=sql.Identifier(column_name),
                    table=sql.Identifier(schema_name, table_name),
                )
            )
            max_id = cursor.fetchone()
            if max_id is None:
                logger.error(
                    f"Error setting primary key for column '{column_name}' in table '{table_name}': no max value found"
                )
                return

            max_id = max_id[0]

            cursor.execute(
                sql.SQL("""
            ALTER TABLE {table}
            ALTER COLUMN {column} SET NOT NULL,
            ALTER COLUMN {column} ADD GENERATED BY DEFAULT AS IDENTITY
              (START WITH {start_with}),
            ADD PRIMARY KEY ({column});
            """).format(
                    table=sql.Identifier(schema_name, table_name),
                    column=sql.Identifier(column_name),
                    start_with=sql.Literal(max_id + 1),
                )
            )
            conn.commit()

        except psycopg.Error:
            logger.exception(
                f"Error setting primary key for column '{column_name}' in table '{table_name}'"
            )


def make_columns_primary_key(
    uri: str,
    *,
    schema_name: str = "public",
    table_name: str,
    column_names: str | Sequence[str],
):
    """
    Make multiple columns as primary key but without auto increment (identity).

    This is similar to make_columns_unique() but with primary key constraint.
    """
    with psycopg.connect(
        conninfo=uri,
    ) as conn:
        try:
            cursor = conn.cursor()

            if isinstance(column_names, str):
                column_names = [column_names]

            cursor.execute(
                sql.SQL("""
                    ALTER TABLE {table}
                    ADD PRIMARY KEY ({columns});
                """).format(
                    table=sql.Identifier(schema_name, table_name),
                    columns=sql.SQL(",").join(
                        sql.Identifier(col) for col in column_names
                    ),
                )
            )
            conn.commit()

        except psycopg.Error:
            logger.exception(
                f"Error setting primary key for column '{column_names}' in table '{table_name}'"
            )


def make_columns_unique(
    uri: str,
    *,
    schema_name: str = "public",
    table_name: str,
    column_names: str | Sequence[str],
):
    """
    Set unique constraint on a column or columns in a table.

    If multiple columns are provided, the unique constraint will be on the combination of the columns.
    """
    with psycopg.connect(
        conninfo=uri,
    ) as conn:
        try:
            cursor = conn.cursor()

            if isinstance(column_names, str):
                column_names = [column_names]

            cursor.execute(
                query=sql.SQL("""
                    ALTER TABLE {table}
                    ADD CONSTRAINT {table_unique_constraint}
                      UNIQUE ({columns});
                """).format(
                    table=sql.Identifier(schema_name, table_name),
                    table_unique_constraint=sql.Identifier(
                        f"{schema_name}-{table_name}-{'-'.join(column_names)}-unique_constraint"
                    ),
                    columns=sql.SQL(",").join(
                        sql.Identifier(col) for col in column_names
                    ),
                )
            )
            conn.commit()

        except psycopg.Error:
            logger.exception(
                f"Error setting unique constraint for column '{column_names}' in table '{table_name}'"
            )


def make_large_columns_unique(
    uri: str,
    *,
    schema_name: str = "public",
    table_name: str,
    column_names: str | Sequence[str],
):
    """
    Use this when the values are large texts, e.g. fasta sequences.

    Reference:
        - https://stackoverflow.com/questions/71379137/how-to-solve-postgresql-index-width-problem
    """
    with psycopg.connect(
        conninfo=uri,
    ) as conn:
        try:
            cursor = conn.cursor()

            if isinstance(column_names, str):
                column_names = [column_names]

            cursor.execute(
                query=sql.SQL("""
                CREATE UNIQUE INDEX ON {table} (
                  {columns}
                );
                """).format(
                    table=sql.Identifier(schema_name, table_name),
                    columns=sql.SQL(",").join(
                        sql.SQL("md5(") + sql.Identifier(col) + sql.SQL(")")
                        for col in column_names
                    ),
                )
            )
            conn.commit()

        except psycopg.Error:
            logger.exception(
                f"Error setting unique index for column '{column_names}' in table '{table_name}'"
            )


def split_column_str_to_list(
    uri: str,
    *,
    schema_name: str = "public",
    table_name: str,
    in_column: str,
    out_column: str,
    separator: str,
    pg_element_type: str = "text",
):
    """
    Split a string column into a list column.
    """
    if pg_element_type.lower() not in {
        "text",
    }:
        raise ValueError(f"Unsupported PostgreSQL element type: {pg_element_type}")

    list_type = sql.SQL(f"{pg_element_type}[]")  # type: ignore

    with psycopg.connect(
        conninfo=uri,
    ) as conn:
        try:
            cursor = conn.cursor()

            # split the string into a list, and write it to a new column
            # plus remove the old column
            cursor.execute(
                query=sql.SQL("""
                ALTER TABLE {table}
                ADD COLUMN {out_column} {list_type};
                """).format(
                    table=sql.Identifier(schema_name, table_name),
                    out_column=sql.Identifier(out_column),
                    list_type=list_type,
                )
            )
            cursor.execute(
                query=sql.SQL("""
                UPDATE {table}
                SET {out_column} = STRING_TO_ARRAY({in_column}, %s)::{list_type};
                """).format(
                    table=sql.Identifier(schema_name, table_name),
                    out_column=sql.Identifier(out_column),
                    in_column=sql.Identifier(in_column),
                    list_type=list_type,
                ),
                params=(separator,),
            )

            cursor.execute(
                query=sql.SQL("""
                ALTER TABLE {table}
                DROP COLUMN {in_column};
                """).format(
                    table=sql.Identifier(schema_name, table_name),
                    in_column=sql.Identifier(in_column),
                )
            )
            conn.commit()

        except psycopg.Error:
            logger.exception(
                f"Error splitting column '{in_column}' in table '{table_name}'"
            )


# def polars_write_database(
#     df: pl.DataFrame,
#     *,
#     schema_name: str = "public",
#     table_name: str,
#     connection: str,
#     if_table_exists: DbWriteMode = "fail",
#     engine: DbWriteEngine = "sqlalchemy",
# ):
#     """
#     pl.DataFrame.write_database() but address the issue of writing unsigned and list columns to database.
#
#     Save all list columns as string columns, then split them back to list columns in the database.
#     This takes some time.
#     """
#     # UInt64 to Int64
#     for col in df.columns:
#         if df[col].dtype == pl.UInt64:
#             df = df.with_columns(pl.col(col).cast(pl.Int64).alias(col))
#
#     # List to string
#     columns_with_list = [col for col in df.columns if df[col].dtype == pl.List]
#     col_to_inner_dtype: dict[str, pl.datatypes.DataType] = {}
#     for col in columns_with_list:
#         dtype = df[col].dtype
#         assert isinstance(dtype, pl.List)
#         inner_dtype = dtype.inner
#         assert inner_dtype is not None
#         col_to_inner_dtype[col] = inner_dtype
#
#         df = df.with_columns(
#             pl.col(col)
#             .cast(pl.List(pl.Utf8))
#             .list.join("/@#DSLKF")
#             .alias(f"{col}_strjoinedAOIFDSIUH")
#         )
#         df.drop_in_place(col)
#
#     # Make safe table_name with schema_name.
#     # pl.DataFrame.write_database() only has table_name, so we need to add schema_name.
#     table_with_schema_str = None
#     with psycopg.connect(
#         conninfo=connection,
#     ) as conn:
#         cursor = conn.cursor()
#         table_with_schema_str = (
#             sql.SQL("""{table_name}""")
#             .format(table_name=sql.Identifier(schema_name, table_name))
#             .as_string(cursor)
#         )
#     if table_with_schema_str is None:
#         raise ValueError(
#             "Failed to get table name with schema. Maybe we can't connect to the database."
#         )
#
#     df.write_database(
#         table_name=table_with_schema_str,
#         connection=connection,
#         if_table_exists=if_table_exists,
#         engine=engine,
#     )
#
#     for col in columns_with_list:
#         split_column_str_to_list(
#             uri=connection,
#             schema_name=schema_name,
#             table_name=table_name,
#             in_column=f"{col}_strjoinedAOIFDSIUH",
#             out_column=col,
#             separator="/@#DSLKF",
#             pg_element_type=polars_datatype_to_postgresql_type(col_to_inner_dtype[col]),
#         )


def polars_write_database(
    df: pl.DataFrame,
    *,
    schema_name: str = "public",
    table_name: str,
    connection: str | Engine,
    if_table_exists: DbWriteMode = "fail",
):
    """
    pl.DataFrame.write_database() but address the issue of writing unsigned and list columns to database.

    Reference:
        - https://stackoverflow.com/questions/77098480/polars-psycopg2-write-column-of-lists-to-postgresql
    """
    if isinstance(connection, str):
        connection = create_engine(connection)

    columns_dtype = {col: df[col].dtype for col in df.columns}
    column_name_to_sqlalchemy_type = {
        col: polars_datatype_to_sqlalchemy_type(dtype)
        for col, dtype in columns_dtype.items()
    }

    pd_df = df.to_pandas(use_pyarrow_extension_array=True)

    # If any column has type list[number] in Polars, the pandas DataFrame will have a numpy array.
    # We need to convert it to a list, because `to_sql` doesn't support numpy arrays.
    for col, dtype in columns_dtype.items():
        if isinstance(dtype, pl.List):
            if isinstance(dtype.inner, pl.Utf8):
                continue
            pd_df[col] = pd_df[col].apply(lambda x: x.tolist())

    # ic(pd_df)
    pd_df.to_sql(
        schema=schema_name,
        name=table_name,
        con=connection,
        if_exists=if_table_exists,
        index=False,
        dtype=column_name_to_sqlalchemy_type,  # type: ignore
    )
